package com.gemho.environment.controller;

import cn.hutool.core.date.DateUnit;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.NumberUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.alibaba.fastjson.JSON;
import com.gemho.environment.model.DevicetypeDto;
import com.gemho.environment.model.RelayDto;
import com.gemho.environment.model.SensorData;
import com.gemho.environment.model.SensorDto;
import com.gemho.environment.service.Channelhandler;
import com.gemho.environment.service.DevicetypeService;
import com.gemho.environment.service.SensorService;
import com.gemho.environment.service.mapper.SensorMapper;
import com.gemho.environment.utils.*;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;
import io.netty.channel.group.ChannelGroup;
import io.netty.channel.group.DefaultChannelGroup;
import io.netty.util.concurrent.GlobalEventExecutor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.net.InetSocketAddress;
import java.sql.Timestamp;
import java.text.SimpleDateFormat;
import java.util.*;

import static com.gemho.environment.utils.StringUtil.hexStringToAlgorism;
import static com.gemho.environment.utils.StringUtil.hexStringToString;

@Component
@Slf4j
@ChannelHandler.Sharable
public class ServerHandler extends SimpleChannelInboundHandler<Object> {

    @Autowired
    private ServerToRelayService serverToRelayService;

    @Autowired
    private SensorMapper sensorMapper;

    @Autowired
    private Channelhandler sChannelService;

    @Autowired
    private SensorService sensorService;

    @Autowired
    private DevicetypeService devicetypeService;

    private static ServerHandler serverHandler;

    static ScriptEngine jse = new ScriptEngineManager().getEngineByName("JavaScript");
    //保留所有与服务器建立连接的channel对象，这边的GlobalEventExecutor在写博客的时候解释一下，看其doc
    public static ChannelGroup channelGroup = new DefaultChannelGroup(GlobalEventExecutor.INSTANCE);
    public static Map<String,String> ydMap = new HashMap<>();

    @PostConstruct
    public void init() {
        serverHandler = this;
    }

    @Override
    protected void channelRead0(ChannelHandlerContext ctx, Object msg) throws Exception {
        InetSocketAddress insocket = (InetSocketAddress) ctx.channel().localAddress();
        String clientPort = String.valueOf(insocket.getPort());
        log.info(ctx.channel().remoteAddress() + " " + ctx.channel().id() + " ");

        /*检测类设备-7000端口*/
        if (clientPort.equals(Constants.NETTY_MONITOR_CLIENT_PORT)) {
            log.info("端口号 {} 接收信息: {}", Constants.NETTY_MONITOR_CLIENT_PORT, msg);
            Map<String, Object> rawdataMap = new HashMap();
            rawdataMap = JSON.parseObject(msg.toString(), rawdataMap.getClass()); //上传的原始数据
            log.info("上传的原始数据:{}", rawdataMap);
            String U = rawdataMap.get("U").toString(); //获取CPU编号
            log.info("CPU编号:{}", U);
            SensorDto sensorDto = sensorService.findByDeviceid(U);
            // valid = 1
            if (sensorDto != null && !"valid".equals("0")) {

                if (sensorDto.getExpire() != null && sensorDto.getExpire().before(new Timestamp(System.currentTimeMillis()))) {
                    log.info("传感器设备:" + sensorDto.getDeviceid() + "的到期时间:" + sensorDto.getExpire());
                    ctx.channel().close();
                    ctx.close();
                    return;
                }

                /*配置IP和端口号*/
                if (rawdataMap.containsKey("cfg")) {
                    String cfg = rawdataMap.get("cfg").toString();
                    if (cfg.equals("0")) {
                        ctx.channel().writeAndFlush("{\"ip\":\"" + sensorDto.getIp() + "\"," + "\"port\":" + Integer.valueOf(sensorDto.getPort()) + "}");
                        ctx.channel().close();
                        ctx.close();
                        return;
                    }
                }

                if (rawdataMap.containsKey("t1") && rawdataMap.containsKey("t2")) {
                    int t1 = Integer.valueOf(rawdataMap.get("t1").toString());
                    int t2 = Integer.valueOf(rawdataMap.get("t2").toString());
                    //比较设备时间和设置时间是否一致
                    if (t1 != sensorDto.getTimetrack() || t2 != sensorDto.getTimetrack2()) {
                        Map<String, Object> map = new HashMap<>();
                        map.put("t1", sensorDto.getTimetrack());
                        map.put("t2", sensorDto.getTimetrack2());
                        ctx.channel().writeAndFlush(JsonUtil.jsonObj2Sting(map));
                    }
                }


                /*信号强度*/
                String Q = null;
                if (rawdataMap.containsKey("Q")) {
                    Q = rawdataMap.get("Q").toString();
                    sensorDto.setRssi(Q);
                }

                /*电量信息*/
                Double V = null;
                if (rawdataMap.containsKey("V")) {
                    V = Double.valueOf(rawdataMap.get("V").toString()) / 1000;
                    sensorDto.setBattery(V.toString()); //电池电量
                }

                sensorDto.setOnlinetime(new Timestamp(System.currentTimeMillis()));

                /*更新设备表里的电量和信号和在线时间*/
                sensorService.update(sensorMapper.toEntity(sensorDto));

                /*获取检测项数据*/
                DevicetypeDto type = devicetypeService.findById(Long.valueOf(sensorDto.getDevicetype()));
                String[] key = type.getCode().split(",");
                String[] gs = type.getGs().split(",");
                Long time = System.currentTimeMillis();
                /*循环遍历当前设备检测项code,例如：['TMP','HR','O2']*/
                for (int i = 0; i < key.length; i++) {
                    List<SensorData> datalst = new ArrayList<>();
                    if (rawdataMap.containsKey(key[i])) {
                        //获取上传数据中的检测数据value
                        List<Integer> value = (List<Integer>) rawdataMap.get(key[i]);
                        //如果上传数据不为空，开始处理数据
                        if (value != null && !value.isEmpty()) {
                            for (int j = value.size() - 1; j >= 0; j--) {
                                SensorData sensorData = new SensorData();
                                sensorData.setItem(key[i]);
                                sensorData.setDeviceid(U);
                                //执行公式
                                String zxgs = gs[i].replace("@data", value.get(j).toString());

                                //湿度单独处理，防止超出100
                                if (key[i].equals("HR") || key[i].contains("sHR")) {
                                    try {
                                        if (Double.valueOf(jse.eval(zxgs).toString()) > 100) {
                                            sensorData.setData(String.valueOf(100));
                                        } else {
                                            sensorData.setData(String.valueOf(jse.eval(zxgs)));
                                        }
                                    } catch (ScriptException e) {
                                        e.printStackTrace();
                                    }
                                } else {
                                    try {
                                        sensorData.setData(String.valueOf(jse.eval(zxgs)));
                                    } catch (ScriptException e) {
                                        e.printStackTrace();
                                    }
                                }

                                //时间处理
                                SimpleDateFormat dateformat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                                if (j == value.size() - 1) {
                                    sensorData.setDtime(dateformat.format(time));
                                } else {
                                    int oldTime = (value.size() - 1 - j) * Integer.valueOf(rawdataMap.get("t1").toString());
                                    sensorData.setDtime(dateformat.format(time - oldTime));
                                }

                                if (Q != null) {
                                    sensorData.setRssi(Q);
                                }

                                if (V != null) {
                                    sensorData.setBattery(String.valueOf(V));
                                }

                                log.info("处理后的数据:{}", sensorData);
                                if (j == value.size() - 1) {
//                                warnMsg(sData);
                                    Thread t = new Thread(new Runnable() {
                                        public void run() {
                                            // 起线程执行报警过滤
                                            // warnMsg(sensorData);
                                        }
                                    });
                                    t.start();
                                }
                                datalst.add(sensorData);
                            }

                            Map<String, Object> finalRawdataMap = rawdataMap;
                            Thread t = new Thread(new Runnable() {
                                @SneakyThrows
                                public void run() {
                                    // sendDataToRelay(key, gs, finalRawdataMap, U);
                                    /*设备数据推送到4GLED(如果关联了4GLED)*/
                                    serverToRelayService.sendDataTo4GMonitor(U, msg.toString());
                                }
                            });
                            t.start();

                        } else {
                            log.info("设备{}的检测项为:{},检测项{}没有数据", U, key, key[i]);
                        }
                    } else {
                        log.info("设备{}的检测项为:{},其中当前数据没有'{}'的数据", U, key, key[i]);
                    }
                }
            }
        }

        /*控制器设备对接-8011*/
        if (clientPort.equals(Constants.NETTY_RELAY_CLIENT_PORT)){
            byte[] msgbyte = StringUtil.stringToAscii(String.valueOf(msg));
//            ByteBuf buf = (ByteBuf) msg;
//            byte[] msgbyte = new byte[buf.readableBytes()];
            String rewdata =  StringUtil.bytesToHexString(msgbyte);
            log.info("接收到的设备指令:{}",rewdata);
            String zt = StrUtil.subWithLength(rewdata, 0, 4);
            if (zt.equals("C33C")) {
                /*获取数据长度*/
                int datalength = Integer.parseInt(rewdata.substring(4,6),16);
                log.info("获取数据长度:{}", datalength);
                String CRCData = rewdata.substring(rewdata.length()-4); //获取最后4位
                String ServerCRC = Integer.toHexString(CRC16Util.calcCrc16(msgbyte,3,datalength-2)).toUpperCase();
                ServerCRC = StringUtils.leftPad(ServerCRC, 4, "0").toUpperCase();
                log.info("服务器计算的CRC:{}, 指令的CRC:{}",ServerCRC,CRCData);

                /*获取心跳包数据位数据*/
                int HeartbeatDataNum = datalength-12-2; //心跳包数据位长度
                String Heartbeat = StrUtil.subWithLength(rewdata,30,HeartbeatDataNum*2);
                String deviceId = StrUtil.subWithLength(rewdata, 6, 24);
                RelayDto relayDto = relayService.findByDeviceid(deviceId);
                if (relayDto.getExpire() != null &&relayDto.getExpire().before(new Timestamp(System.currentTimeMillis()))) {
                    log.info("控制器设备:"+relayDto.getDeviceid()+"的到期时间:"+relayDto.getExpire());
                    ctx.channel().close();
                    ctx.close();
                    return;
                }
                /*应答处理*/
                if (rewdata.substring(4,6).equals("0F")) {
                    String ydData = StrUtil.subWithLength(rewdata, 30, 2);
                    log.info("设备{}应答数据:{}",deviceId,ydData);
                    if (Constants.sendMap.containsKey(deviceId)) {
                        if (ydData.equals("00")){
                            ydMap.put(deviceId, "00");
                        }else if(ydData.equals("FF")) {
                            ydMap.put(deviceId, "FF");
                        }
                    }
                }

                /*心跳处理*/
                if (Heartbeat.equals("A55A")) {
                    log.info("接收到设备 {} 的心跳!", deviceId);
                    sChannelService.setDeviceChannel(deviceId,ctx.channel().id().toString());
                    relayService.updateRelayOnlinetime(deviceId);//保存设备在线时间
                    ByteBuf bufff = Unpooled.buffer();
                    bufff.writeBytes(StringUtil.hexString2Bytes(rewdata));//对接需要16进制
                    ctx.channel().writeAndFlush(bufff);
                }

                /*指令处理 start*/
                String commandData = StrUtil.subWithLength(rewdata, 3*2+12*2,2); //获取指令位数据 帧头*2+数据长度*2+设备ID*2

                /*上传继电器状态值指令处理*/
                if (commandData.equals("91")) {
                    String relayStatusData = StrUtil.subWithLength(rewdata, 3*2+12*2+2+4, (datalength-12-1-2-2)*2);
                    log.info("{}上传继电器状态值指令{}",deviceId,relayStatusData);
                    log.info(CRCData);
                    String binData = NumberUtil.getBinaryStr(Integer.parseInt(relayStatusData,16));
                    binData = StringUtils.leftPad(binData, 16, "0");
                    log.info("relayStatusBytes:{}", binData);
                    relayWayService.updateWayStatus(deviceId, binData);
                    //应答
                    ByteBuf bufff = resUtil(CRCData, ServerCRC,deviceId);
                    ctx.channel().writeAndFlush(bufff);
                }else if (commandData.equals("AF")){
                    /*上传工作模式处理*/
                    String workData = StrUtil.subWithLength(rewdata, 3*2+12*2+2+4, (datalength-12-1-2-2)*2);
                    log.info("{}上传工作模式{}",deviceId,workData);
                    if (workData.equals("1111")){
                        relayService.updateRelayWorkMode(deviceId, "1");
                    }else if (workData.equals("0000")){
                        relayService.updateRelayWorkMode(deviceId, "0");
                    }
                    //应答
                    ByteBuf bufff = resUtil(CRCData, ServerCRC,deviceId);
                    ctx.channel().writeAndFlush(bufff);
                }else if (commandData.equals("90")) {
                    log.info("90>>>");
                    /*获取设备传输的检测项*/
                    String deviceIdX = deviceId + "X";
                    String itemNum = StrUtil.subWithLength(rewdata, 3*2+12*2+2, 4);
                    JSONObject jsonObject = JSONUtil.parseObj(RegisterToSensorConstants.REGISTER_To_SENSOR);
                    String item = String.valueOf(jsonObject.get(itemNum)); //寄存器地址
                    String itemDataHex = StrUtil.subWithLength(rewdata, 3*2+12*2+2+4, 8);
                    int itemData = Integer.parseInt(itemDataHex, 16);
                    log.info("设备{}上传监测项为{}的数据为:{}", deviceIdX,item,itemData);
                    Map<String, Object> rawdataMap = new HashMap();
                    rawdataMap.put("U", deviceIdX);
                    rawdataMap.put(item, new int[]{itemData});
                    log.info("jsons数据{}",rawdataMap);

                    SensorDto sensorDto = sensorService.findByDeviceid(deviceIdX);
                    sensorDto.setOnlinetime(new Timestamp(System.currentTimeMillis()));
                    /*更新设备表里的电量和信号和在线时间*/
                    sensorService.update(sensorMapper.toEntity(sensorDto));
                    /*获取检测项数据*/
                    DevicetypeDto type = devicetypeService.findById(Long.valueOf(sensorDto.getDevicetype()));
                    String[] key = type.getCode().split(",");
                    String[] gs = type.getGs().split(",");
                    Long time =System.currentTimeMillis();

                    SensorData sensorData = new SensorData();
                    sensorData.setItem(item);
                    sensorData.setDeviceid(deviceIdX);

                    for(int i = 0; i<key.length; i++){
                        List<SensorData> datalst = new ArrayList<>();
                        if (rawdataMap.containsKey(key[i])) {
                            List<Integer> value = new ArrayList<>();
                            value.add(itemData);
                            if (value != null && !value.isEmpty()) {
                                for (int j = value.size() -1 ; j >= 0; j--){

                                    sensorData.setItem(key[i]);
                                    sensorData.setDeviceid(deviceIdX);
                                    //执行公式
                                    String zxgs = gs[i].replace("@data", value.get(j).toString());

                                    //湿度单独处理，防止超出100
                                    if (key[i].equals("HR") || key[i].contains("sHR")){
                                        try {
                                            if (Double.valueOf(jse.eval(zxgs).toString()) > 100) {
                                                sensorData.setData(String.valueOf(100));
                                            }else {
                                                sensorData.setData(String.valueOf(jse.eval(zxgs)));
                                            }
                                        } catch (ScriptException e) {
                                            e.printStackTrace();
                                        }
                                    } else {
                                        try {
                                            sensorData.setData(String.valueOf(jse.eval(zxgs)));
                                        } catch (ScriptException e) {
                                            e.printStackTrace();
                                        }
                                    }

                                    SimpleDateFormat dateformat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                                    if (j == value.size() -1) {
                                        sensorData.setDtime(dateformat.format(time));
                                    }else {
                                        int oldTime = (value.size()-1-j) * Integer.valueOf(rawdataMap.get("t1").toString());
                                        sensorData.setDtime(dateformat.format(time-oldTime));
                                    }

                                    log.info("处理后的数据:{}",sensorData);
                                    if (j == value.size() - 1 ) {
                                        Thread t = new Thread(new Runnable(){
                                            public void run(){
                                                // 起线程执行报警过滤
                                                warnMsg(sensorData);
                                            }});
                                        t.start();
                                    }
                                    datalst.add(sensorData);
                                }
                                //写入mangoDB
                                mongoTemplate.insertAll(datalst);
                            }
                        }
                    }


                    //应答
                    ByteBuf bufff = resUtil(CRCData, ServerCRC,deviceId);
                    ctx.channel().writeAndFlush(bufff);
                }
            }
        }

        if (clientPort.equals(Constants.NETTY_MIST_CLIENT_PORT)){
            String receivedata =  msg.toString();
            log.info("端口:{},外采烟感报警器发送数据:{}",Constants.NETTY_MIST_CLIENT_PORT, receivedata);
            String type = receivedata.substring(12,14);//获取消息类型
            log.info("消息类型:{}", type);
            if ("02".equals(type)) { //接收到心跳包
                log.info("外采烟感报警器发送心跳包");
                ctx.channel().writeAndFlush("2b633273000b02004f4bda");
            }else if ("04".equals(type)){ //设备发送当前状态
                log.info("外采烟感报警器发送自身消息");
                ctx.channel().writeAndFlush("2b633273000b04004f4bdc");
            }else if ("0a".equals(type)){
                log.info("外采烟感报警器发送复位消息");
                ctx.channel().writeAndFlush("2b633273000b09004f4be1");
            }else if ("09".equals(type)){
                log.info("外采烟感报警器发送报警消息");
                int length = hexStringToAlgorism(receivedata.substring(8,12));//接收到的数据字节数
                String equipid = hexStringToString(receivedata.substring(16,20),2) + receivedata.substring(20,32);//主设备
                int total = hexStringToAlgorism(receivedata.substring(32,34));//报警总数
                String defence = hexStringToString(receivedata.substring(58,64),2);//报警防区
                String code = hexStringToString(receivedata.substring(48,54),2);//报警代码
                log.info("信息长度：" + length + ":" + receivedata.length());
                log.info("主设备ID：" + equipid);
                log.info("报警总数：" + total);
                log.info("报警防区：" + defence);
                log.info("报警代码：" + code);
                if ("111".equals(code)){
                    ctx.channel().writeAndFlush("2b633273000b09004f4be1");
                    if (!smokeAlarmCache.checkSmokeAlarmCache(equipid)){
                        log.info("设备{},10分钟内没有报警",equipid);
                        smokeAlarmCache.setSmokeAlarmCache(equipid);
                        //保存报警记录
                        Warnmsg sWarnmsg = new Warnmsg();
                        sWarnmsg.setDeviceid(equipid);
                        sWarnmsg.setItem("SW");
                        sWarnmsg.setFlg("0");
                        sWarnmsg.setMessage("发生烟雾报警");
                        sWarnmsg.setNt("true");
                        warnmsgService.create(sWarnmsg);
                        //获取设备关联的用户
                        List<User> users = userSensorService.getUserBySensorId(equipid);
                        //发送短信报警
                        for (User user : users) {
                            SDKTestSendTemplateSMS.sendMsg(user.getPhone(), "487466", new String[]{equipid, "发生烟雾报警"});
                        }
                    }else {
                        log.info("设备{},10分钟内有报警",equipid);
                    }
                }
            }
        }
    }


    //表示服务端与客户端连接建立
    @Override
    public void handlerAdded(ChannelHandlerContext ctx) throws Exception {
        Channel channel = ctx.channel();  //其实相当于一个connection
        /* 添加 */
        channelGroup.add(channel);
    }

    @Override
    public void handlerRemoved(ChannelHandlerContext ctx) throws Exception {
        Channel channel = ctx.channel();

        //验证一下每次客户端断开连接，连接自动地从channelGroup中删除调。
        log.info("{}断开连接", channel.remoteAddress());
        try {
            sChannelService.rmDeviceChannel(channel.id().toString());
        }catch (Exception e) {

        }

        //删除连接池的通道
        channelGroup.remove(channel);
    }

    //连接处于活动状态
    @Override
    public void channelActive(ChannelHandlerContext ctx) throws Exception {
        Channel channel = ctx.channel();
        log.info(channel.remoteAddress() +" 上线了");
    }

    @Override
    public void channelInactive(ChannelHandlerContext ctx) throws Exception {
        Channel channel = ctx.channel();
        log.info(channel.remoteAddress() +" 下线了");
        channelGroup.remove(channel);
    }

    @Override
    public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception {
        Channel channel = ctx.channel();
        cause.printStackTrace();
        channelGroup.remove(channel);
        ctx.close();
    }

    @Override
    public void userEventTriggered (ChannelHandlerContext ctx, Object evt) throws Exception {
        Channel channel = ctx.channel();
        log.info("{}心跳超时,删除本次连接",channel.remoteAddress());
        try {
            sChannelService.rmDeviceChannel(channel.id().toString());
        }catch (Exception e) {

        }

        //删除连接池的通道
        channelGroup.remove(channel);
        ctx.close();
    }
}
